@@ -34,19 +34,11 @@
 #include <glib.h>
 #include <locale.h>
 
-#ifdef ENABLE_GNOME_KEYRING
-#include <gnome-keyring.h>
-#endif
-
 #include "defines.h"
 #include "../config.h"
 
 #include "gksu-context.h"
 
-#ifdef ENABLE_GNOME_KEYRING
-gboolean keyring_used = FALSE;
-#endif
-
 /* local function declarations */
 static void
 gksu_context_init (GTypeInstance *instance, gpointer g_class);
@@ -173,15 +165,19 @@ gksu_context_get_command (GksuContext *context)
 void
 gksu_context_set_password (GksuContext *context, gchar *password)
 {
-  g_assert (password != NULL);
-
   if (context->password)
-    g_free (context->password);
+    {
+      g_free (context->password);
+      context->password = NULL;
+    }
 
-  if (password[strlen(password) - 1] == '\n')
-    context->password = g_strdup (password);
-  else
-    context->password = g_strdup_printf ("%s\n", password);
+  if (password && strcmp ("", password) && strcmp ("\n", password))
+    {
+      if (password[strlen(password) - 1] == '\n')
+	context->password = g_strdup (password);
+      else
+	context->password = g_strdup_printf ("%s\n", password);
+    }
 }
 
 /**
@@ -376,8 +372,6 @@ gksu_context_dispose (GObject *object)
   g_free (self->user);
   g_free (self->password);
   g_free (self->command);
-
-  g_free (self->user);
 }
 
 static void
@@ -417,6 +411,30 @@ gksu_context_error_get_type (void)
   return etype;
 }
 
+static gchar *
+get_process_name (pid_t pid)
+{
+#ifdef linux
+  FILE *f;
+  char *cmdline = malloc (sizeof(gchar*)*256);
+  char *cmdfile = g_strdup_printf("/proc/%i/status",pid);
+
+  cmdline[255] = '\0';
+
+  f = fopen(cmdfile, "r");
+  g_free (cmdfile);
+  if(f != NULL) 
+    {
+      fgets(cmdline, 255, f);
+      cmdline[255] = '\0';
+      fclose(f);
+    }
+  return cmdline;
+#else
+  return NULL;
+#endif
+}
+
 /**
  * prepare_xauth:
  *
@@ -512,9 +530,9 @@ copy (const char *fn, const char *dir)
     {
       if (errno == EEXIST)
 	fprintf (stderr,
-		   _("The X authority file i am trying to create for"
-		     " the target user already exists! This is highly"
-		     " suspicous!"));
+		 _("Impossible to create the .Xauthority file: a file "
+		   "already exists. This might be a security issue; "
+		   "please investigate or warn your system administrator."));
       else
 	fprintf (stderr,
 		   _("Error copying '%s' to '%s': %s"),
@@ -601,78 +619,6 @@ sudo_reset_xauth (GksuContext *context, gchar *xauth,
   g_free (xauth);
 }
 
-#ifdef ENABLE_GNOME_KEYRING
-gboolean
-try_gnome_keyring_password (GksuContext *context)
-{
-  GnomeKeyringAttributeList *attributes;
-  GnomeKeyringAttribute attribute;
-  GnomeKeyringResult result;
-  GList *list;
-  GError *error = NULL;
-  gboolean keyring_has_password;
-  gchar *keyring_password;
-  gchar *keyring_command;
-
-  keyring_used = TRUE;
-
-  attributes = gnome_keyring_attribute_list_new ();
-
-  attribute.name = g_strdup ("user");
-  attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
-  attribute.value.string = g_strdup (gksu_context_get_user (context));
-  g_array_append_val (attributes, attribute);
-
-  attribute.name = g_strdup ("type");
-  attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
-  attribute.value.string = g_strdup ("local");
-  g_array_append_val (attributes, attribute);
-
-  attribute.name = g_strdup ("creator");
-  attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
-  attribute.value.string = g_strdup ("gksu");
-  g_array_append_val (attributes, attribute);
-      
-  list = g_list_alloc();
-  keyring_has_password = FALSE;
-
-  result = gnome_keyring_find_items_sync (GNOME_KEYRING_ITEM_GENERIC_SECRET,
-					  attributes,
-					  &list);
-  gnome_keyring_attribute_list_free (attributes);
-  if (
-      (result == GNOME_KEYRING_RESULT_OK) &&
-      (g_list_length(list) == 1)
-      )
-    {
-      GnomeKeyringFound *found = list->data;
-      keyring_password = g_strdup(found->secret);
-
-      keyring_command = g_strdup (gksu_context_get_command (context));
-      gksu_context_set_command (context, "/bin/echo test > /dev/null");
-      gksu_context_set_password (context, keyring_password);
-      gksu_context_run (context, &error);
-
-      if (!error)
-	keyring_has_password = TRUE;
-      gksu_context_set_command (context, keyring_command);
-    }
-
-  if (keyring_has_password) 
-    {
-      gchar *password;
-
-      password = g_locale_from_utf8 (keyring_password, 
-				     strlen (keyring_password), 
-				     NULL, NULL, NULL);
-      gksu_context_set_password (context, password);
-      return FALSE; /* we do not need a password */
-    }
-
-  return TRUE;
-}
-#endif
-
 gboolean
 try_su_run (GksuContext *context)
 {
@@ -727,6 +673,7 @@ try_su_run (GksuContext *context)
       fd_set rfds;
 
       struct timeval tv;
+      struct termios tio;
 
       gchar buf[256];
       gint status;
@@ -744,14 +691,14 @@ try_su_run (GksuContext *context)
       if (context->debug)
 	fprintf (stderr, "try_su_run: buf: -%s-\n", buf);
 
-      if (!strncmp (buf, "Password: ", 10))
-	{
-	  write (fdpty, "\n", 1);
-	  close (fdpty);
-	}
-      else
+      usleep (100);
+
+      tcgetattr (fdpty, &tio);
+      
+      if ((tio.c_lflag & ECHO))
 	need_pass = FALSE;
 
+      kill (pid, SIGKILL);
       waitpid (pid, &status, 0);
     }
 
@@ -764,6 +711,7 @@ try_su_run (GksuContext *context)
  *
  * Checks if we need to ask for a password or if we have ways of
  * getting the password for ourselves or we simply don't need it.
+ * Since 1.3.0.
  *
  * Returns: TRUE if requesting a password is needed, FALSE otherwise.
  *
@@ -776,30 +724,38 @@ gksu_context_try_need_password (GksuContext *context)
   if (try_su_run (context) == FALSE)
     return FALSE;
 
-#ifdef ENABLE_GNOME_KEYRING
-  if (try_gnome_keyring_password (context) == FALSE)
-    return FALSE;
-#endif
-
   return TRUE;
 }
 
 /**
- * gksu_context_run:
+ * gksu_context_run_full:
  * @context: a #GksuContext
+ * @ask_pass: a #GksuAskPasswordFunc
+ * @user_data: a #gpointer with user data to be passed to the
+ * #GksuAskPasswordFunc
+ * @pass_not_needed: a #GksuPasswordNotNeededFunc
+ * @pnn_user_data: a #gpointer with the user data to be passed to the
+ * #GksuPasswordNotNeededFunc
  * @error: a #GError object to be filled with the error code or NULL
  *
  * This could be considered one of the main functions in GKSu.
- * it is responsible for doing the 'user changing' magic.
+ * it is responsible for doing the 'user changing' magic calling
+ * the #GksuAskPasswordFunc function to request a password if needed.
+ * and the #GksuPasswordNotNeededFunc function if a password won't be
+ * needed, so the application has the oportunity of warning the user
+ * what it's doing.
+ * Since 1.3.3.
  *
  * Returns: the child's error status, 0 if all went fine, -1 if failed
  */
 gboolean
-gksu_context_run (GksuContext *context, GError **error)
+gksu_context_run_full (GksuContext *context, 
+		       GksuAskPasswordFunc ask_pass, gpointer user_data, 
+		       GksuPasswordNotNeededFunc pass_not_needed, 
+		       gpointer pnn_user_data, GError **error)
 {
   GQuark gksu_quark;
   int i = 0;
-  gboolean auth_failed = FALSE;
 
   gchar auxcommand[] = PREFIX "/lib/" PACKAGE "/gksu-run-helper";
 
@@ -882,15 +838,13 @@ gksu_context_run (GksuContext *context, GError **error)
       gint target_uid = -1;
       gint my_uid = 0;
 
-      gchar buf[256];
+      gchar buf[256] = {0};
       gint status;
 
-      /* 
-	 for controling the conversation with the
-	 helper
-      */
-      gboolean first_line_read = FALSE;
-      
+      gchar *cmdline = NULL;
+
+      gboolean password_needed = FALSE;
+
       my_uid = getuid();
       pwd = getpwnam (context->user);
       if (pwd)
@@ -899,17 +853,100 @@ gksu_context_run (GksuContext *context, GError **error)
       /* no need to ask for password if we're already root */
       if (my_uid != target_uid && my_uid)
 	{
-	  if (context->password)
+	  if (context->password || ask_pass)
 	    {
+	      struct termios tio;
+
+	      usleep (100);
 	      read (fdpty, buf, 256);
 	      if (context->debug)
 		fprintf (stderr, "gksu_context_run: buf: -%s-\n", buf);
-	      sleep (1);
-	      write (fdpty, context->password, 
-		     strlen(context->password));
+
+	      usleep (100);
+
+	      tcgetattr (fdpty, &tio);
+
+	      if (!(tio.c_lflag & ECHO))
+		{
+		  if (!context->password)
+		    {
+		      if (ask_pass (context, buf, user_data, error))
+			  return -1;
+
+		      g_return_val_if_fail (context->password, -1);
+		    }
+
+		  usleep (100);
+
+		  password_needed = TRUE;
+		  write (fdpty, context->password, 
+			 strlen(context->password));
+		}
 	    }
 	}
+
+      if (!password_needed)
+	pass_not_needed (context, pnn_user_data);
       
+      if (context->debug)
+	fprintf (stderr, "DEBUG (run:after-pass) buf: -%s-\n", buf);
+      if (strncmp (buf, "gksu", 4) && strncmp (buf, "su", 2))
+	{
+	  bzero (buf, 256);
+	  /* drop the \n echoed on password entry if su did request
+	     a password */
+	  if (password_needed)
+	    read (fdpty, buf, 256);
+	  read (fdpty, buf, 255);
+	  if (context->debug)
+	    fprintf (stderr, "DEBUG (run:post-after-pass) buf: -%s-\n", buf);
+	}
+
+      if (!strncmp (buf, "su: Authentication failure", 26))
+	{
+	  g_set_error (error, gksu_quark, 
+		       GKSU_CONTEXT_ERROR_WRONGPASS,
+		       _("Wrong password."));
+	  if (context->debug)
+	    fprintf (stderr, "DEBUG (auth_failed) buf: -%s-\n", buf);
+	}
+      else if (!strncmp (buf, "gksu: waiting", 13))
+	{
+	  gchar *line;
+	  
+	  if (context->debug)
+	    fprintf (stderr, "DEBUG (gksu: waiting) buf: -%s-\n", buf);
+
+	  line = g_strdup_printf ("gksu-run: %s\n", context->display);
+	  write (fdpty, line, strlen(line));
+	  g_free (line);
+	  
+	  line = g_strdup_printf ("gksu-run: %s\n", context->xauth);
+	  write (fdpty, line, strlen(line));
+	  g_free (line);
+
+	  tcdrain (fdpty);
+
+	  bzero (buf, 256);
+	  read (fdpty, buf, 256);
+	}
+      else
+	{
+	  gchar *emsg = g_strdup_printf (_("Failed to communicate with "
+					   "gksu-run-helper.\n\n"
+					   "Received:\n"
+					   " %s\n"
+					   "While expecting:\n"
+					   " %s"), buf, "gksu: waiting");
+	  g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_HELPER, emsg);
+	  g_free (emsg);
+
+	  if (context->debug)
+	    fprintf (stderr, "DEBUG (failed!) buf: -%s-\n", buf);
+
+	  return -1;
+	}
+
       FD_ZERO(&rfds);
       FD_SET(fdpty, &rfds);
 
@@ -918,6 +955,15 @@ gksu_context_run (GksuContext *context, GError **error)
 
       while (!waitpid (pid, &status, WNOHANG))
 	{
+	  bzero (buf, 256);
+	  usleep (200);
+
+	  if (cmdline)
+	    g_free (cmdline);
+	  cmdline = get_process_name (pid);
+
+	  usleep(1000);
+
 	  if (select (fdpty+1, &rfds, NULL, NULL, &tv) < 0)
 	    {
 	      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_PIPEREAD,
@@ -931,107 +977,10 @@ gksu_context_run (GksuContext *context, GError **error)
 	    {
 	      bzero(buf, 256);
 	      read (fdpty, buf, 255);
-
-	      if (!first_line_read)
-		{
-		  /* we've already read something important on the
-		     outside of the loop, probably because su gave
-		     no output at all (pam_wheel trust or something)
-		  */
-		  if (strncmp (buf, "gksu", 4))
-		    read (fdpty, buf, 255);
-
-		  if (!strncmp (buf, "su: Authentication failure", 26))
-		    auth_failed = TRUE;
-		  else if (!strncmp (buf, "gksu: waiting", 13))
-		    {
-		      gchar *line;
-		      
-		      line = g_strdup_printf ("gksu-run: %s\n", context->display);
-		      write (fdpty, line, strlen(line));
-		      g_free (line);
-		      
-		      line = g_strdup_printf ("gksu-run: %s\n", context->xauth);
-		      write (fdpty, line, strlen(line));
-		      g_free (line);
-
-		      bzero (buf, 255);
-		    }
-
-#ifdef ENABLE_GNOME_KEYRING
-		  if (keyring_used && context->password)
-		    {
-		      static GMainLoop *keyring_loop = NULL;
-		      GnomeKeyringAttributeList *attributes;
-		      GnomeKeyringAttribute attribute;
-		      
-		      gchar *key_name;
-		      
-		      attributes = gnome_keyring_attribute_list_new ();
-		      
-		      attribute.name = g_strdup ("user");
-		      attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
-		      attribute.value.string = g_strdup (gksu_context_get_user (context));
-		      g_array_append_val (attributes, attribute);
-		    
-		      attribute.name = g_strdup ("type");
-		      attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
-		      attribute.value.string = g_strdup ("local");
-		      g_array_append_val (attributes, attribute);
-
-		      attribute.name = g_strdup ("creator");
-		      attribute.type = GNOME_KEYRING_ATTRIBUTE_TYPE_STRING;
-		      attribute.value.string = g_strdup ("gksu");
-		      g_array_append_val (attributes, attribute);
-      
-		      key_name = g_strdup_printf ("Local password for user %s",
-						  gksu_context_get_user (context));
-      
-		      keyring_loop = g_main_loop_new (NULL, FALSE);
-
-		      void
-			keyring_create_item_cb (GnomeKeyringResult result,
-						guint32 id, gpointer data)
-			{
-			  g_main_loop_quit (keyring_loop);
-			}
-
-		      gnome_keyring_item_create (NULL,
-						 GNOME_KEYRING_ITEM_GENERIC_SECRET,
-						 key_name,
-						 attributes,
-						 gksu_context_get_password (context),
-						 TRUE,
-						 keyring_create_item_cb, NULL, NULL);
-		      gnome_keyring_attribute_list_free (attributes);
-		      g_main_loop_run (keyring_loop);
-		    }
-#endif
-
-		  first_line_read = TRUE;
-		}
-	      else if (!strncmp (buf, "gksu-run:", 8))
-		{
-		  gchar *tmp = NULL;
-
-		  tmp = g_strrstr (buf, "\n");
-		  while (!tmp)
-		    {
-		      bzero(buf, 256);
-		      read (fdpty, buf, 255);
-		      tmp = g_strrstr (tmp, "\n");
-		    }
-		  if (tmp)
-		    fprintf (stderr, "%s", tmp+1);
-		  bzero (buf, 255);
-		}
-	      else
-		fprintf (stderr, "%s", buf);
+	      fprintf (stderr, "%s", buf);
 	    }
 	  else
 	    FD_SET(fdpty, &rfds);
-
-	  usleep (200);
 	}
 
       while (1)
@@ -1046,19 +995,63 @@ gksu_context_run (GksuContext *context, GError **error)
 	{
 	  if (WEXITSTATUS(status))
 	    {
-	      if (auth_failed)
-		g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_WRONGPASS,
-			     _("Wrong password."));
-	      else
-		g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_CHILDFAILED,
-			     _("Child terminated with %d status"),
+	      if(cmdline)
+		{
+		  /* su already exec()ed something else, don't report
+		   * exit status errors in that case
+		   */
+		  if (!g_str_has_suffix (cmdline, "su\n"))
+		    {
+		      g_free (cmdline);
+		      return TRUE;
+		    }
+		  g_free (cmdline);
+		}
+
+	      if (error == NULL)
+		g_set_error (error, gksu_quark, 
+			     GKSU_CONTEXT_ERROR_CHILDFAILED,
+			     _("su terminated with %d status"),
 			     WEXITSTATUS(status));
-	      return WEXITSTATUS(status);
 	    }
 	}
     }
 
-  return 0;
+  if (error)
+    return TRUE;
+
+  return FALSE;
+}
+
+/* DO NOT USE -- Deprecated */
+gboolean
+gksu_context_ask_and_run (GksuContext *context, GksuAskPasswordFunc ask_pass, 
+			  gpointer user_data, GError **error)
+{
+  return gksu_context_run_full (context, ask_pass, user_data, NULL, NULL, error);
+}
+
+/**
+ * gksu_context_run:
+ * @context: a #GksuContext
+ * @error: a #GError object to be filled with the error code or NULL
+ *
+ * This could be considered one of the main functions in GKSu.
+ * it is responsible for doing the 'user changing' magic.
+ *
+ * Returns: the child's error status, 0 if all went fine, -1 if failed
+ */
+gboolean
+gksu_context_run (GksuContext *context, GError **error)
+{
+  return gksu_context_run_full (context, NULL, NULL, NULL, NULL, error);
+}
+
+static gboolean
+ack_killed_child (gpointer pid)
+{
+      waitpid (GPOINTER_TO_INT(pid), NULL, 0);
+      return FALSE;
 }
 
 static gboolean
@@ -1066,10 +1059,9 @@ try_sudo_validation (GksuContext *context)
 {
   char **cmd;
   int argcount = 4;
-  char buffer[256];
+  char buffer[256] = {0};
 
   pid_t pid;
-  int status;
   size_t r;
   FILE *infile, *outfile;
   int parent_pipe[2];	/* For talking to the parent */
@@ -1077,8 +1069,6 @@ try_sudo_validation (GksuContext *context)
 
   gboolean need_pass = TRUE;
   
-  bzero(buffer, 256);
-
   if ((pipe(parent_pipe)) == -1)
     return TRUE;
 
@@ -1155,7 +1145,6 @@ try_sudo_validation (GksuContext *context)
 	    fprintf (stderr, "We'll need a password...\n");
 
 	  fprintf (outfile, "\n");
-	  fclose (outfile);
 	} 
       else 
 	{
@@ -1164,24 +1153,10 @@ try_sudo_validation (GksuContext *context)
 	  need_pass = FALSE;
 	}
 
+      /* politely request that sudo GET OUT OF THE WAY */
+      kill (pid, SIGKILL);
 
-      while (!waitpid (pid, &status, WNOHANG))
-	{
-	  bzero(buffer, 256);
-	  if(!fgets (buffer, 255, infile))
-	    break;
-	  fprintf (stderr, "%s", buffer);
-	}
-
-      /* make sure we did read everything */
-      while (1)
-	{
-	  bzero(buffer, 256);
-	  if(!fread (buffer, sizeof(gchar), 255, infile))
-	    break;
-	  fprintf (stderr, "%s", buffer);
-	  fflush (stderr);
-	}
+      g_timeout_add (3000, ack_killed_child, GINT_TO_POINTER(pid));
     }
 
   return need_pass;
@@ -1193,6 +1168,7 @@ try_sudo_validation (GksuContext *context)
  *
  * Checks if we need to ask for a password or if we have ways of
  * getting the password for ourselves or we simply don't need it.
+ * Since 1.3.1.
  *
  * Returns: TRUE if requesting a password is needed, FALSE otherwise.
  *
@@ -1207,22 +1183,34 @@ gksu_context_sudo_try_need_password (GksuContext *context)
 }
 
 /**
- * gksu_context_sudo_run:
+ * gksu_context_sudo_run_full:
  * @context: a #GksuContext
+ * @ask_pass: a #GksuAskPasswordFunc
+ * @user_data: a #gpointer with user data to be passed to the
+ * #GksuAskPasswordFunc
+ * @pass_not_needed: a #GksuPasswordNotNeededFunc
+ * @pnn_user_data: a #gpointer with the user data to be passed to the
+ * #GksuPasswordNotNeededFunc
  * @error: a #GError object to be filled with the error code or NULL
  *
  * This could be considered one of the main functions in GKSu.
- * it is responsible for doing the 'user changing' magic by
- * calling gksu_ask_password() if it needs the user's password
- * it behaves like sudo.
+ * it is responsible for doing the 'user changing' magic calling
+ * the #GksuAskPasswordFunc function to request a password if needed.
+ * and the #GksuPasswordNotNeededFunc function if a password won't be
+ * needed, so the application has the oportunity of warning the user
+ * what it's doing.
+ * Since 1.3.3.
  *
  * Returns: the child's error status, 0 if all went fine, -1 if failed
  */
 gboolean
-gksu_context_sudo_run (GksuContext *context, GError **error)
+gksu_context_sudo_run_full (GksuContext *context, 
+			    GksuAskPasswordFunc ask_pass, gpointer user_data, 
+			    GksuPasswordNotNeededFunc pass_not_needed, 
+			    gpointer pnn_user_data, GError **error)
 {
   char **cmd;
-  char buffer[256];
+  char buffer[256] = {0};
   int argcount = 8;
   int i, j;
 
@@ -1311,28 +1299,49 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
       argcount++;
     }
 
-  for (i = j = 0; ; i++)
-    {
-      if (context->command[i] == ' ' || context->command[i] == '\0')
-	{
-	  buffer[j] = '\0';
-	  cmd = g_realloc (cmd, sizeof(gchar*) * (argcount + 1));
-	  cmd[argcount] = g_strdup (buffer);
-	  bzero (buffer, 256);
-	  argcount = argcount + 1;
-	  j = 0;
+  {
+    gchar *tmp_arg = g_malloc (sizeof(gchar)*1);
+    gboolean inside_quotes = FALSE;
 
-	  if (context->command[i] == '\0')
-	    break;
-	}
-      else
-	{
-	  if (context->command[i] == '\\')
+    tmp_arg[0] = '\0';
+
+    for (i = j = 0; ; i++)
+      {
+	if ((context->command[i] == '\'') && (context->command[i-1] != '\\'))
+	  {
 	    i = i + 1;
-	  buffer[j] = context->command[i];
-	  j = j + 1;
-	}
-    }
+	    inside_quotes = !inside_quotes;
+	  }
+
+	if ((context->command[i] == ' ' && inside_quotes == FALSE) 
+	    || context->command[i] == '\0')
+	  {
+	    tmp_arg = g_realloc (tmp_arg, sizeof(gchar)*(j+1));
+	    tmp_arg[j] = '\0';
+	    cmd = g_realloc (cmd, sizeof(gchar*) * (argcount + 1));
+	    cmd[argcount] = g_strdup (tmp_arg);
+
+	    g_free (tmp_arg);
+
+	    argcount = argcount + 1;
+	    j = 0;
+
+	    if (context->command[i] == '\0')
+	      break;
+
+	    tmp_arg = g_malloc (sizeof(gchar)*1);
+	    tmp_arg[0] = '\0';
+	  }
+	else
+	  {
+	    if (context->command[i] == '\\')
+	      i = i + 1;
+	    tmp_arg = g_realloc (tmp_arg, sizeof(gchar)*(j+1));
+	    tmp_arg[j] = context->command[i];
+	    j = j + 1;
+	  }
+      }
+  }
   cmd = g_realloc (cmd, sizeof(gchar*) * (argcount + 1));
   cmd[argcount] = NULL;
 
@@ -1387,7 +1396,7 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
     } 
   else 
     {
-      gboolean auth_failed = FALSE;
+      gchar *cmdline = NULL;
 
       // Parent
       close(parent_pipe[1]);
@@ -1427,7 +1436,8 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
 	  if (context->debug)
 	    fprintf (stderr, "Yeah, we're in...\n");
 
-	  if (context->password == NULL || (!strcmp (context->password, "")))
+	  if ((context->password == NULL || (!strcmp (context->password, "")))
+	      && ask_pass == NULL)
 	    {
 	      g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_NOPASSWORD,
 			   _("No password was supplied and sudo needs it."));
@@ -1436,6 +1446,16 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
 	      return -1;
 	    }
 
+	  if (!context->password)
+	    {
+	      if (ask_pass (context, _("Password: "), user_data, error))
+		return -1;
+	      
+	      g_return_val_if_fail (context->password, -1);
+	    }
+	  
+	  usleep (100);
+
 	  fprintf (outfile, "%s", context->password);
 	  fclose (outfile);
 
@@ -1444,20 +1464,51 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
 	  */
 	  fgets (buffer, 255, infile);
 	  fgets (buffer, 255, infile);
-	  if (!strcmp (buffer, "Sorry, try again.\n"))
-	    auth_failed = TRUE;
 	} 
       else 
 	{
 	  if (context->debug)
-	    fprintf (stderr, "Oops... what's up?\n");
+	    fprintf (stderr, "No password prompt found; we'll assume we don't need a password.\n");
+
+	  if (pass_not_needed)
+	    pass_not_needed (context, pnn_user_data);
 
 	  fprintf (stderr, "%s", buffer);
 	}
 
+      if (!strcmp (buffer, "Sorry, try again.\n"))
+	g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_WRONGPASS,
+		     _("Wrong password."));
+      else if (!strncmp (buffer, "Sorry, user ", 12))
+	g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_NOT_ALLOWED,
+		     _("The underlying authorization mechanism (sudo) "
+		       "does not allow you to run this program. Contact "
+		       "the system administrator."));
+      else
+	{
+	  gchar *haystack = buffer;
+	  gchar *needle;
+
+	  needle = g_strstr_len (haystack, strlen (haystack), " ");
+	  if (needle && (needle + 1))
+	    {
+	      needle += 1;
+	      if (!strncmp (needle, "is not in", 9))
+		g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_NOT_ALLOWED,
+			     _("The underlying authorization mechanism (sudo) "
+			       "does not allow you to run this program. Contact "
+			       "the system administrator."));
+	    }
+	}
 
       while (!waitpid (pid, &status, WNOHANG))
 	{
+	  if (cmdline)
+	    g_free (cmdline);
+	  cmdline = get_process_name (pid);
+
+	  usleep(1000);
+
 	  bzero(buffer, 256);
 	  if(!fgets (buffer, 255, infile))
 	    break;
@@ -1480,19 +1531,57 @@ gksu_context_sudo_run (GksuContext *context, GError **error)
 	{
 	  if (WEXITSTATUS(status))
 	    {
-	      if (auth_failed)
-		g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_WRONGPASS,
-			     _("Wrong password."));
-	      else
-		g_set_error (error, gksu_quark, GKSU_CONTEXT_ERROR_CHILDFAILED,
-			     _("Child terminated with %d status"),
-			     WEXITSTATUS(status));
+	      if(cmdline)
+		{
+		  /* sudo already exec()ed something else, don't report
+		   * exit status errors in that case
+		   */
+		  if (!g_str_has_suffix (cmdline, "sudo\n"))
+		    {
+		      g_free (cmdline);
+		      return TRUE;
+		    }
+		  g_free (cmdline);
+		}
 
-	      return TRUE;
+	      if (error == NULL)
+		g_set_error (error, gksu_quark, 
+			     GKSU_CONTEXT_ERROR_CHILDFAILED,
+			     _("sudo terminated with %d status"),
+			     WEXITSTATUS(status));
 	    }
 	}
     }
   
+  /* if error is set we have found an error condition */
+  if (error)
+    return TRUE;
+
   return FALSE;
 }
 
+/* DO NOT USE -- Deprecated */
+gboolean
+gksu_context_sudo_ask_and_run (GksuContext *context, GksuAskPasswordFunc ask_pass, 
+			       gpointer user_data, GError **error)
+{
+  return gksu_context_sudo_run_full (context, ask_pass, user_data, NULL, NULL, error);
+}
+
+/**
+ * gksu_context_sudo_run:
+ * @context: a #GksuContext
+ * @error: a #GError object to be filled with the error code or NULL
+ *
+ * This could be considered one of the main functions in GKSu.
+ * it is responsible for doing the 'user changing' magic by
+ * calling gksu_ask_password() if it needs the user's password
+ * it behaves like sudo.
+ *
+ * Returns: the child's error status, 0 if all went fine, -1 if failed
+ */
+gboolean
+gksu_context_sudo_run (GksuContext *context, GError **error)
+{
+  return gksu_context_sudo_run_full (context, NULL, NULL, NULL, NULL, error);
+}
